use std::collections::{HashMap, HashSet};

use proc_macro2::{Ident, Span};

use btrun::End;

use crate::{
    error::MaybeErr,
    parsing::{DestinationState, GetSpan, Message, State},
    Protocol,
};

impl Protocol {
    pub(crate) fn validate(&self) -> syn::Result<()> {
        let validator = ProtocolValidator::new(self);
        validator
            .all_states_declared_and_used()
            .combine(validator.receivers_and_senders_matched())
            .combine(validator.no_undeliverable_msgs())
            .combine(validator.replies_expected())
            .combine(validator.msg_sent_or_received())
            .combine(validator.clients_only_receive_replies())
            .into()
    }
}

struct ProtocolValidator<'a> {
    protocol: &'a Protocol,
    actors_by_states: HashMap<&'a Ident, &'a Ident>,
    client_actors: HashSet<&'a Ident>,
}

impl<'a> ProtocolValidator<'a> {
    fn new(protocol: &'a Protocol) -> Self {
        let mut actors_by_states: HashMap<&Ident, &Ident> = HashMap::new();
        for actor_def in protocol.actor_defs.iter() {
            for state in actor_def.states.as_ref().iter() {
                actors_by_states.insert(state, &actor_def.actor);
            }
        }
        let client_actors: HashSet<&Ident> = protocol
            .transitions
            .iter()
            .filter(|transition| transition.in_msg().is_none())
            .map(|transition| actors_by_states.get(&transition.in_state.state_trait))
            .filter(|option| option.is_some())
            .map(|option| *option.unwrap())
            .collect();
        Self {
            protocol,
            actors_by_states,
            client_actors,
        }
    }

    /// Returns the [Ident] for the actor that the given state is a part of.
    fn actor(&self, state: &Ident) -> Option<&Ident> {
        self.actors_by_states.get(state).copied()
    }

    fn is_client(&self, actor: &Ident) -> bool {
        self.client_actors.contains(actor)
    }

    fn is_client_state(&self, state: &State) -> bool {
        self.actor(&state.state_trait)
            .map(|actor| self.is_client(actor))
            .unwrap_or(false)
    }

    const UNDECLARED_STATE_ERR: &str = "State was not declared.";
    const UNUSED_STATE_ERR: &str = "State was declared but never used.";

    /// Verifies that every state which is used has been declared, except for the End state.
    fn all_states_declared_and_used(&self) -> MaybeErr {
        let end = Ident::new(End::ident(), Span::call_site());
        let mut declared: HashSet<&Ident> = HashSet::new();
        declared.insert(&end);
        for actor_def in self.protocol.actor_defs.iter() {
            for state in actor_def.states.as_ref().iter() {
                declared.insert(state);
            }
        }
        let mut used: HashSet<&Ident> = HashSet::with_capacity(declared.len());
        for transition in self.protocol.transitions.iter() {
            let in_state = &transition.in_state;
            used.insert(&in_state.state_trait);
            used.extend(in_state.owned_states.as_ref().iter());
            if let Some(in_msg) = transition.in_msg() {
                used.extend(in_msg.owned_states.as_ref().iter());
            }
            for out_states in transition.out_states.as_ref().iter() {
                used.insert(&out_states.state_trait);
                used.extend(out_states.owned_states.as_ref().iter());
            }
            // We don't have to check the states referred to in out_msgs because the
            // receivers_and_senders_matched method ensures that each of these exists in a receiver
            // position.
        }
        let undeclared: MaybeErr = used
            .difference(&declared)
            .map(|ident| syn::Error::new(ident.span(), Self::UNDECLARED_STATE_ERR))
            .collect();
        let unused: MaybeErr = declared
            .difference(&used)
            .filter(|ident| **ident != End::ident())
            .map(|ident| syn::Error::new(ident.span(), Self::UNUSED_STATE_ERR))
            .collect();
        undeclared.combine(unused)
    }

    const UNMATCHED_SENDER_ERR: &str = "No receiver found for message type.";
    const UNMATCHED_RECEIVER_ERR: &str = "No sender found for message type.";

    /// Ensures that the recipient state for every sent message has a receiving transition
    /// defined, and every receiver has a sender.
    fn receivers_and_senders_matched(&self) -> MaybeErr {
        let mut senders: HashSet<(&State, &Message)> = HashSet::new();
        let mut receivers: HashSet<(&State, &Message)> = HashSet::new();
        for transition in self.protocol.transitions.iter() {
            if let Some(msg) = transition.in_msg() {
                receivers.insert((&transition.in_state, msg));
            }
            for dest in transition.out_msgs.as_ref().iter() {
                let dest_state = match &dest.state {
                    DestinationState::Individual(dest_state) => dest_state,
                    DestinationState::Service(dest_state) => dest_state,
                };
                senders.insert((dest_state, &dest.msg));
            }
        }
        let extra_senders: MaybeErr = senders
            .difference(&receivers)
            .map(|pair| syn::Error::new(pair.1.msg_type.span(), Self::UNMATCHED_SENDER_ERR))
            .collect();
        let extra_receivers: MaybeErr = receivers
            .difference(&senders)
            .map(|pair| syn::Error::new(pair.1.msg_type.span(), Self::UNMATCHED_RECEIVER_ERR))
            .collect();
        extra_senders.combine(extra_receivers)
    }

    const UNDELIVERABLE_ERR: &str =
        "Receiver must either be a service, an owned state, or an out state, or the message must be a reply.";

    /// Checks that messages are only sent to destinations which are either services, states
    /// which are owned by the sender, listed in the output states, or that the message is a
    /// reply.
    fn no_undeliverable_msgs(&self) -> MaybeErr {
        let mut err = MaybeErr::none();
        for transition in self.protocol.transitions.iter() {
            let mut allowed_states: Option<HashSet<&Ident>> = None;
            for dest in transition.out_msgs.as_ref().iter() {
                if dest.msg.is_reply() {
                    continue;
                }
                match &dest.state {
                    DestinationState::Service(_) => continue,
                    DestinationState::Individual(dest_state) => {
                        let allowed = allowed_states.get_or_insert_with(|| {
                            transition
                                .out_states
                                .as_ref()
                                .iter()
                                .map(|state| &state.state_trait)
                                .chain(transition.in_state.owned_states.as_ref().iter())
                                .collect()
                        });
                        if !allowed.contains(&dest_state.state_trait) {
                            err = err.combine(
                                syn::Error::new(
                                    dest_state.state_trait.span(),
                                    Self::UNDELIVERABLE_ERR,
                                )
                                .into(),
                            );
                        }
                    }
                }
            }
        }
        err
    }

    const INVALID_REPLY_ERR: &str =
        "Replies can only be used in transitions which handle messages.";
    const MULTIPLE_REPLIES_ERR: &str =
        "Only a single reply can be sent in response to any message.";

    /// Verifies that exactly one reply is sent in response to a previously sent message.
    fn replies_expected(&self) -> MaybeErr {
        let mut err = MaybeErr::none();
        for transition in self.protocol.transitions.iter() {
            let replies: Vec<_> = transition
                .out_msgs
                .as_ref()
                .iter()
                .map(|dest| &dest.msg)
                .filter(|msg| msg.is_reply())
                .collect();
            if replies.is_empty() {
                continue;
            }
            if replies.len() > 1 {
                err = err.combine(
                    replies
                        .iter()
                        .map(|reply| {
                            syn::Error::new(reply.msg_type.span(), Self::MULTIPLE_REPLIES_ERR)
                        })
                        .collect(),
                );
            }
            if transition.in_msg().is_none() {
                err = err.combine(
                    replies
                        .iter()
                        .map(|reply| {
                            syn::Error::new(reply.msg_type.span(), Self::INVALID_REPLY_ERR)
                        })
                        .collect(),
                );
            }
        }
        err
    }

    const NO_MSG_SENT_OR_RECEIVED_ERR: &str = "A transition must send or receive a message.";

    /// Verifies that either a message is received, or sent by a transition. The rational behind
    /// this is that if no message is sent or received, then the state transition is unobserved by
    /// other actors, and so should not be in a protocol.
    fn msg_sent_or_received(&self) -> MaybeErr {
        self.protocol
            .transitions
            .iter()
            .filter(|transition| {
                transition.in_msg().is_none() && transition.out_msgs.as_ref().is_empty()
            })
            .map(|transition| syn::Error::new(transition.span(), Self::NO_MSG_SENT_OR_RECEIVED_ERR))
            .collect()
    }

    const CLIENT_RECEIVED_NON_REPLY_ERR: &str =
        "A client actor cannot receive a message which is not a reply.";

    /// A client is any actor with a state that sends at least one message when not handling an
    /// incoming message. Such actors are not allowed to receive any message which are not replies.
    fn clients_only_receive_replies(&self) -> MaybeErr {
        self.protocol
            .transitions
            .iter()
            .filter(|transition| {
                if let Some(msg) = transition.in_msg() {
                    if !msg.is_reply() {
                        return self.is_client_state(&transition.in_state);
                    }
                }
                false
            })
            .map(|transition| {
                syn::Error::new(transition.span(), Self::CLIENT_RECEIVED_NON_REPLY_ERR)
            })
            .collect()
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::parsing::{ActorDef, Dest, NameDef, Transition};

    macro_rules! assert_ok {
        ($maybe_err:expr) => {
            let result: syn::Result<()> = $maybe_err.into();
            assert!(result.is_ok(), "{}", result.err().unwrap());
        };
    }

    macro_rules! assert_err {
        ($maybe_err:expr, $expected_msg:expr) => {
            let result: syn::Result<()> = $maybe_err.into();
            assert!(result.is_err());
            assert_eq!($expected_msg, result.err().unwrap().to_string());
        };
    }

    /// Creates minimal [Protocol] value.
    fn min_protocol() -> Protocol {
        const STATE_NAME: &str = "Init";
        Protocol::new(
            NameDef::new("Test"),
            [
                ActorDef::new("server", [STATE_NAME]),
                ActorDef::new("client", ["Client"]),
            ],
            [
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("End", [])],
                    [Dest::new(
                        DestinationState::Service(State::new(STATE_NAME, [])),
                        Message::new("Msg", false, []),
                    )],
                ),
                Transition::new(
                    State::new(STATE_NAME, []),
                    Some(Message::new("Msg", false, [])),
                    [State::new("End", [])],
                    [],
                ),
            ],
        )
    }

    #[test]
    fn all_states_declared_and_used_ok() {
        let protocol = min_protocol();
        let result = ProtocolValidator::new(&protocol).all_states_declared_and_used();

        assert_ok!(result);
    }

    #[test]
    fn all_states_declared_and_used_end_not_used_ok() {
        const STATE_NAME: &str = "Init";
        let protocol = Protocol::new(
            NameDef::new("Test"),
            [ActorDef::new("actor", [STATE_NAME])],
            [Transition::new(
                State::new(STATE_NAME, []),
                Some(Message::new("Activate", false, [])),
                [State::new(STATE_NAME, [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.all_states_declared_and_used();

        assert_ok!(result);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_err() {
        let protocol = Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("Next", [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.all_states_declared_and_used();

        assert_err!(result, ProtocolValidator::UNDECLARED_STATE_ERR);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_out_state_owned_err() {
        let protocol = Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init", "Next"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("Init", []), State::new("Next", ["Undeclared"])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.all_states_declared_and_used();

        assert_err!(result, ProtocolValidator::UNDECLARED_STATE_ERR);
    }

    #[test]
    fn all_states_declared_and_used_undeclared_in_state_owned_err() {
        let protocol = Protocol::new(
            NameDef::new("Undeclared"),
            [ActorDef::new("actor", ["Init", "Next"])],
            [Transition::new(
                State::new("Init", ["Undeclared"]),
                Some(Message::new("Activate", false, [])),
                [State::new("Next", [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.all_states_declared_and_used();

        assert_err!(result, ProtocolValidator::UNDECLARED_STATE_ERR);
    }

    #[test]
    fn all_states_declared_and_used_unused_err() {
        let protocol = Protocol::new(
            NameDef::new("Unused"),
            [ActorDef::new("actor", ["Init", "Extra"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("End", [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.all_states_declared_and_used();

        assert_err!(result, ProtocolValidator::UNUSED_STATE_ERR);
    }

    #[test]
    fn receivers_and_senders_matched_ok() {
        let protocol = min_protocol();
        let result = ProtocolValidator::new(&protocol).receivers_and_senders_matched();

        assert_ok!(result);
    }

    #[test]
    fn receivers_and_senders_matched_unmatched_sender_err() {
        let protocol = Protocol::new(
            NameDef::new("Unbalanced"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                None,
                [State::new("Init", [])],
                [Dest::new(
                    DestinationState::Service(State::new("Init", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.receivers_and_senders_matched();

        assert_err!(result, ProtocolValidator::UNMATCHED_SENDER_ERR);
    }

    #[test]
    fn receivers_and_senders_matched_unmatched_receiver_err() {
        let protocol = Protocol::new(
            NameDef::new("Unbalanced"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("NotExists", false, [])),
                [State::new("Init", [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.receivers_and_senders_matched();

        assert_err!(result, ProtocolValidator::UNMATCHED_RECEIVER_ERR);
    }

    #[test]
    fn no_undeliverable_msgs_ok() {
        let protocol = min_protocol();
        let result = ProtocolValidator::new(&protocol).no_undeliverable_msgs();

        assert_ok!(result);
    }

    #[test]
    fn no_undeliverable_msgs_reply_ok() {
        let protocol = Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Listening", "Client"])],
            [Transition::new(
                State::new("Listening", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Listening", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Client", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.no_undeliverable_msgs();

        assert_ok!(result);
    }

    #[test]
    fn no_undeliverable_msgs_service_ok() {
        let protocol = Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Service(State::new("Server", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.no_undeliverable_msgs();

        assert_ok!(result);
    }

    #[test]
    fn no_undeliverable_msgs_owned_ok() {
        let protocol = Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["FileClient", "FileHandle"])],
            [Transition::new(
                State::new("FileClient", ["FileHandle"]),
                None,
                [State::new("FileClient", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("FileHandle", [])),
                    Message::new("FileOp", false, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.no_undeliverable_msgs();

        assert_ok!(result);
    }

    #[test]
    fn no_undeliverable_msgs_err() {
        let protocol = Protocol::new(
            NameDef::new("Undeliverable"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Server", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.no_undeliverable_msgs();

        assert_err!(result, ProtocolValidator::UNDELIVERABLE_ERR);
    }

    #[test]
    fn replies_expected_ok() {
        let protocol = Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Server", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Server", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Client", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.replies_expected();

        assert_ok!(result);
    }

    #[test]
    fn replies_expected_invalid_reply_err() {
        let protocol = Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "Server"])],
            [Transition::new(
                State::new("Client", []),
                None,
                [State::new("Client", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Server", [])),
                    Message::new("Msg", true, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.replies_expected();

        assert_err!(result, ProtocolValidator::INVALID_REPLY_ERR);
    }

    #[test]
    fn replies_expected_multiple_replies_err() {
        let protocol = Protocol::new(
            NameDef::new("ValidReplies"),
            [ActorDef::new("actor", ["Client", "OtherClient", "Server"])],
            [Transition::new(
                State::new("Server", []),
                Some(Message::new("Msg", false, [])),
                [State::new("Server", [])],
                [
                    Dest::new(
                        DestinationState::Individual(State::new("Client", [])),
                        Message::new("Msg", true, []),
                    ),
                    Dest::new(
                        DestinationState::Individual(State::new("OtherClient", [])),
                        Message::new("Msg", true, []),
                    ),
                ],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.replies_expected();

        assert_err!(result, ProtocolValidator::MULTIPLE_REPLIES_ERR);
    }

    #[test]
    fn msg_sent_or_received_msg_received_ok() {
        let protocol = Protocol::new(
            NameDef::new("Test"),
            [ActorDef::new("actor", ["Init"])],
            [Transition::new(
                State::new("Init", []),
                Some(Message::new("Activate", false, [])),
                [State::new("End", [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.msg_sent_or_received();

        assert_ok!(result);
    }

    #[test]
    fn msg_sent_or_received_msg_sent_ok() {
        let protocol = Protocol::new(
            NameDef::new("Test"),
            [ActorDef::new("actor", ["First", "Second"])],
            [Transition::new(
                State::new("First", []),
                None,
                [State::new("First", [])],
                [Dest::new(
                    DestinationState::Individual(State::new("Second", [])),
                    Message::new("Msg", false, []),
                )],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.msg_sent_or_received();

        assert_ok!(result);
    }

    #[test]
    fn msg_sent_or_received_neither_err() {
        let protocol = Protocol::new(
            NameDef::new("Test"),
            [ActorDef::new("actor", ["First"])],
            [Transition::new(
                State::new("First", []),
                None,
                [State::new("First", [])],
                [],
            )],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.msg_sent_or_received();

        assert_err!(result, ProtocolValidator::NO_MSG_SENT_OR_RECEIVED_ERR);
    }

    #[test]
    fn clients_only_receive_replies_ok() {
        let protocol = Protocol::new(
            NameDef::new("ClientReplies"),
            [
                ActorDef::new("client", ["Client", "Waiting"]),
                ActorDef::new("server", ["Init", "Listening"]),
            ],
            [
                Transition::new(
                    State::new("Init", []),
                    Some(Message::new("Activate", false, [])),
                    [State::new("Listening", [])],
                    [],
                ),
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Waiting", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Listening", [])),
                        Message::new("Ping", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Listening", []),
                    Some(Message::new("Ping", false, [])),
                    [State::new("Listening", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Waiting", [])),
                        Message::new("Ping", true, []),
                    )],
                ),
                Transition::new(
                    State::new("Waiting", []),
                    Some(Message::new("Ping", true, [])),
                    [State::new("Client", [])],
                    [],
                ),
            ],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.clients_only_receive_replies();

        assert_ok!(result);
    }

    #[test]
    fn clients_only_receive_replies_err() {
        let protocol = Protocol::new(
            NameDef::new("ClientReplies"),
            [
                ActorDef::new("client", ["Client", "Waiting"]),
                ActorDef::new("server", ["Init", "Listening"]),
            ],
            [
                Transition::new(
                    State::new("Init", []),
                    Some(Message::new("Activate", false, [])),
                    [State::new("Listening", [])],
                    [],
                ),
                Transition::new(
                    State::new("Client", []),
                    None,
                    [State::new("Waiting", [])],
                    [Dest::new(
                        DestinationState::Service(State::new("Listening", [])),
                        Message::new("Ping", false, []),
                    )],
                ),
                Transition::new(
                    State::new("Listening", []),
                    Some(Message::new("Ping", false, [])),
                    [State::new("Listening", [])],
                    [Dest::new(
                        DestinationState::Individual(State::new("Waiting", [])),
                        Message::new("Ping", true, []),
                    )],
                ),
                Transition::new(
                    State::new("Waiting", []),
                    Some(Message::new("Ping", true, [])),
                    [State::new("Client", [])],
                    [],
                ),
                Transition::new(
                    State::new("Client", []),
                    Some(Message::new("Activate", false, [])),
                    [State::new("Client", [])],
                    [],
                ),
            ],
        );
        let input = ProtocolValidator::new(&protocol);

        let result = input.clients_only_receive_replies();

        assert_err!(result, ProtocolValidator::CLIENT_RECEIVED_NON_REPLY_ERR);
    }
}
